/*
 *	C P . C
 *
 *	Copyright (C) 1988-1991 John W Grothman.  All rights reserved.
 *
 *	purpose:
 *	  CodPr source code printer. CodPr adds a left margin and line number to
 *	  each line in the source file and a header to each page.
 *
 *	to do:
 *
 *	technical notes:
 *	  CodPr prints source files with no more than 80 characters per line at
 *	  12 cpi and 8 lpi.
 *
 *	modification log:
 *	1.0		4/8/88		John W. Grothman
 *			8/19		jwg - fix RESET code to add 'start 6 LPI' code
 *	1.1		10/4		jwg - require infile and outfile args; outfile may
 *							  be "prn", "lptx" (x=1-3), "con" or a filename
 *	1.2		12/22		jwg - allow multiple file names; add -o parm to modify
 *							  the default outfile (LPT1)
 *	1.3		12/28		jwg - CodPr now prints the file's last modification time
 *							  and date in the page header instead of the
 *							  current time and date
 *	1.4		1/17/89		jwg - replace hard tabs with spaces using TABSIZE
 *							  instead of letting the printer handle tabs
 *	2.0		4/18		jwg - change MAXLPP from 80 to 84
 *							- change header text from Time/Date to Last Mod
 *	2.1		8/4			jwg - change '?' parm to '-h' (for 'help')
 *	3.0		1/15/90		jwg - generalize code for PD release
 *	3.1		5/22		jwg - mods to support the HP LaserJet III
 *	4.0		10/14		jwg - port to Turbo C++
 *							- explicitly close 'outfile' - oops
 *	4.1		2/18/91		jwg - mods to wrap long lines without incrementing
 *							  linenum instead of writing off the end of page
 *	4.2		2/27		jwg - merge HPLJ3 and EPSON code into one executable;
 *							  default to HPLJ3; switch to EPSON with new -e
 *							  command line parameter
 */

#include <stdio.h>
#include <dos.h>
#include <fcntl.h>
#include <conio.h>
#include <io.h>
#include <jwg.h>

#define MAXCPL	256					/* maximum line length */

#define HDRLEN	4					/* number of lines in page header */
#define HDRMRG	16					/* 16 spaces (page header margin) */
#define MARGIN	9					/* 9 spaces (line margin) */
#define LINLEN	78					/* line length minus 2 corners */

#define	TABSIZE	4					/* 4 colums per hard tab */

#define ULC 	''					/* upper left-hand corner */
#define URC 	''					/* upper right-hand corner */
#define VLS 	''					/* vertical line segment */
#define HLS		''					/* horizontal line segment */
#define LLC 	''					/* lower left-hand corner */
#define LRC 	''					/* lower right-hand corner */

/* EPSON definitions */
#define EMAXLPP	84							/* maximum lines per page */
char	einit[] = "\033\072\033\060";			/* 12 CPI, 8 LPI */
char	ereset[] = "\022\033\101\014\033\062";	/* 10 CPI, 6 LPI, start 6 LPI */

/* HPLJ3 definitions */
#define LMAXLPP	78
char	linit[] = "\033\046\154\070\104\033\046\153\064\123";
char	lreset[] = "\033\046\154\066\104\033\046\153\060\123";

int		maxlpp = LMAXLPP;					/* set defaults to HPLJ3 */
char	*init = linit;
char	*reset = lreset;

int		hour, minute, second, day, month, year;
char	line[MAXCPL - (HDRMRG * 2)];		/* times 2 for safety */
char	bufr[MAXCPL];

int		main(int, char **);
void	PrintHeader(FILE *, int, char *);
void	Usage(void);

int main(argc, argv)
  int			argc;
  char			**argv;
{
  FILE			*infile;
  FILE			*outfile = NULL;
  int			i, j, k, len, nargs;
  int			linenum, pagenum, charnum;
  char			*fname;
  char			ch;
  char			*chp;
  int			fh;
  int			col;
  int			tab;
  struct ftime	ftd;

	if (argc < 2)
	{
		Usage();
		_exit(FAILURE);
	}

	nargs = 1;
	tab = TABSIZE;

	while (argv[nargs][0] == '-')
	{
		if (argv[nargs][1] == 'o')
		{
			if (!(outfile = fopen(&(argv[nargs][2]), "w")))
			{
				cprintf	(
							"\r\nError: can't open (%s) for output\r\n",
							&(argv[nargs][2])
						);
				Usage();
				_exit(FAILURE);
			}
			++nargs;
		}
		else if (argv[nargs][1] == 't')
		{
			if ((tab = atoi(&(argv[nargs][2]))) == 0)
			{
				cprintf	(
							"\r\nError: (%s) isn't a valid tabsize\r\n",
							&(argv[nargs][2])
						);
				Usage();
				_exit(FAILURE);
			}
			++nargs;
		}
		else if (argv[nargs][1] == 'e')
		{
			maxlpp = EMAXLPP;		/* switch codes to EPSON */
			init = einit;
			reset = ereset;
			++nargs;
		}
		else if (argv[nargs][1] == 'h')
		{
			Usage();
			_exit(FAILURE);
		}
		else
		{
			cprintf("\r\nError: invalid option\r\n");
			Usage();
			_exit(FAILURE);
		}
	}

	/* if no output filename specified on the command line (-ofn) */
	if (!outfile)
	{
		if (!(outfile = fopen("LPT1", "w")))
		{
			cprintf("\r\nError: can't open (LPT1) for output\r\n");
			_exit(FAILURE);
		}
	}

	cprintf("\r\nCodPr - Version 4.2\r\n\n");		/* 2/27/91 */

	while (argv[nargs] != NULL)
	{
		if (!(infile = fopen(argv[nargs], "r")))
		{
			cprintf("\r\nError: can't open (%s) for input\r\n", argv[nargs]);
			++nargs;
			continue;
		}

		fh = _open(argv[nargs], O_RDONLY);
		getftime(fh, &ftd);
		_close(fh);

		hour = ftd.ft_hour;
		minute = ftd.ft_min;
		second = ftd.ft_tsec * 2;
		month = ftd.ft_month;
		day = ftd.ft_day;
		year = ftd.ft_year + 80;

		fprintf(outfile, init);

		chp = argv[nargs];
		while (*chp != 0)
			++chp;
		while ((ch = *chp) != '\\' && ch != ':' && chp != argv[nargs])
			--chp;
		if (chp != argv[nargs])
			fname = ++chp;
		else
			fname = chp;

		cprintf("    Printing: %-12s    Page #     ", fname);

		linenum = 1;
		pagenum = 1;
		do
		{
			cprintf("\b\b\b%-3u", pagenum);
			PrintHeader(outfile, pagenum, fname);
			for (i = 0; i < maxlpp - HDRLEN; ++i)
			{
				if (fgets(line, MAXCPL, infile) == NULL)
					break;
				for (j = 0, col = 0; j < strlen(line); j++)
				{
					if ((ch = line[j]) == '\t')
					{
						do
						{
							bufr[col++] = ' ';
						} while ((col & (tab - 1)) != 0);
					}
					else
					{
						bufr[col++] = ch;
					}
				}
				bufr[col] = NULL;
				for (j = 0; j < MARGIN; j++)
					fputc(' ', outfile);
				fprintf(outfile, "%4u:  ", linenum);
				k = 0;
				charnum = 0;
				while ((ch = bufr[k++]) != '\n')
				{
					if (charnum < (LINLEN + 2))
					{
						fputc(ch, outfile);
						charnum++;
					}
					else
					{
						fputc('\n', outfile);
						for (j = 0; j < MARGIN - 2; j++)
							fputc(' ', outfile);
						fprintf(outfile, "C %4u:  ", linenum);
						fputc(ch, outfile);
						charnum = 1;
						i++;		/* bump loop counter to cover extra line */
					}
				}
				fputc(ch, outfile);
				++linenum;
			}
			fputc('\f', outfile);	/* do a formfeed */
			++pagenum;
		} while (!feof(infile));

		fclose(infile);

		fprintf(outfile, reset);
		fflush(outfile);

		fputc('\n', stdout);
		++nargs;
	}

	fclose(outfile);
	return(SUCCESS);
}

/* Prints a rectangle 80 columns wide by 3 lines long using the IBM box drawing
 * characters. Prints the current filename and page number.
 */
void PrintHeader(ld, p, fname)
  FILE		*ld; 			/* list device */
  int		p;				/* page number */
  char		*fname;			/* file name */
{
  int		i;

	for (i = 0; i < HDRMRG; i++)
		fputc(' ', ld);
	fputc(ULC, ld);					/* upper left-hand corner */
	for (i = 0; i < LINLEN; ++i)	/* long horizontal line segment */
		fputC(HLS, ld);
	fputc(URC, ld);					/* upper rignt-hand corner */
	fputc('\n', ld);
	for (i = 0; i < HDRMRG; i++)
		fputc(' ', ld);
	fputc(VLS, ld);					/* vertical line segment */
	fprintf(ld, "  File:  %-12s", fname);
	fprintf(ld, "        Revision:  %02d:%02d:%02d", hour, minute, second);
	fprintf(ld, "  %02d/%02d/%02d", month, day, year);
	fprintf(ld, "        Page:  %-3u  ", p);
	fputc(VLS, ld);
	fputc('\n', ld);
	for (i = 0; i < HDRMRG; i++)
		fputc(' ', ld);
	fputc(LLC, ld);					/* lower left-hand corner */
	for (i = 0; i < LINLEN; ++i)
		fputc(HLS, ld);
	fputc(LRC, ld);					/* lower right-hand corner */
	fputs("\n\n\n", ld);
}

/* Prints usage/help message
 */

void Usage()
{
	cprintf("\r\n  usage:  CP [-ofn] [-tn] [-e] [-h] file [file]\r\n");
	cprintf("\r\n  where:  -ofn    fn = output filename    (default = LPT1)");
	cprintf("\r\n          -tn     n = tab stop increment  (default = 4)");
	cprintf("\r\n          -e      use EPSON control codes (default = HPLJ3)");
	cprintf("\r\n          -h      print usage and exit");
	cprintf("\r\n          file    one or more filenames   (wildcards OK)\r\n");
	cprintf("\r\n  note:   All options must precede the first filename!");
	cprintf("\r\n");
	return;
}
